from types import SimpleNamespace

from common import (
    make_mistral3_picollama,
    make_qwen2vl_picollama,
    make_tokenizer,
    run_and_check_merge,
)
from transformers import AutoConfig, Qwen2Config, Qwen2ForCausalLM

from mergekitty.architecture import get_architecture_info
from mergekitty.config import (
    InputSliceDefinition,
    MergeConfiguration,
    OutputSliceDefinition,
)
from mergekitty.merge import _model_out_config


class TestQwen2VLArchitecture:
    def test_qwen2_vl_architecture_info_expands_vision_and_text_weights(self, tmp_path):
        model_path = make_qwen2vl_picollama(tmp_path / "qwen2vl")
        make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)

        cfg = AutoConfig.from_pretrained(model_path)
        arch_info = get_architecture_info(cfg)
        weights = arch_info.all_weights(cfg)
        names = {weight.name for weight in weights}

        assert arch_info.name() == "qwen2_vl"
        assert arch_info.num_layers(cfg) == 2
        assert len(weights) == 58
        assert "visual.blocks.1.attn.qkv.weight" in names
        assert "visual.merger.mlp.2.bias" in names
        assert "model.layers.1.self_attn.q_proj.weight" in names
        assert "model.norm.weight" in names
        assert "lm_head.weight" in names

    def test_qwen2_vl_passthrough_stack_updates_nested_text_layer_count(self, tmp_path):
        model_path = make_qwen2vl_picollama(tmp_path / "qwen2vl")
        make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)

        config = MergeConfiguration(
            merge_method="passthrough",
            slices=[
                OutputSliceDefinition(
                    sources=[
                        InputSliceDefinition(model=model_path, layer_range=[0, 2]),
                    ]
                ),
                OutputSliceDefinition(
                    sources=[
                        InputSliceDefinition(model=model_path, layer_range=[0, 2]),
                    ]
                ),
            ],
            dtype="bfloat16",
        )

        def _check_config_layers(path: str):
            out_cfg = AutoConfig.from_pretrained(path)
            assert out_cfg.text_config.num_hidden_layers == 4

        run_and_check_merge(config, validate=_check_config_layers)


class TestMistral3Architecture:
    def test_mistral3_architecture_info_expands_vision_and_text_weights(self, tmp_path):
        model_path = make_mistral3_picollama(tmp_path / "mistral3")
        make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)

        cfg = AutoConfig.from_pretrained(model_path)
        arch_info = get_architecture_info(cfg)
        weights = arch_info.all_weights(cfg)
        names = {weight.name for weight in weights}

        assert arch_info.name() == "mistral3"
        assert arch_info.definition.match_model_type == "mistral"
        assert arch_info.num_layers(cfg) == 2
        assert len(weights) == 47
        assert "vision_tower.transformer.layers.1.attention.q_proj.weight" in names
        assert "multi_modal_projector.patch_merger.merging_layer.weight" in names
        assert "language_model.model.layers.1.self_attn.q_proj.weight" in names
        assert "language_model.model.norm.weight" in names
        assert "language_model.lm_head.weight" in names

    def test_mistral3_passthrough_stack_updates_nested_text_layer_count(self, tmp_path):
        model_path = make_mistral3_picollama(tmp_path / "mistral3")
        make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)

        config = MergeConfiguration(
            merge_method="passthrough",
            slices=[
                OutputSliceDefinition(
                    sources=[
                        InputSliceDefinition(model=model_path, layer_range=[0, 2]),
                    ]
                ),
                OutputSliceDefinition(
                    sources=[
                        InputSliceDefinition(model=model_path, layer_range=[0, 2]),
                    ]
                ),
            ],
            dtype="bfloat16",
        )

        def _check_config_layers(path: str):
            out_cfg = AutoConfig.from_pretrained(path)
            assert out_cfg.text_config.num_hidden_layers == 4

        run_and_check_merge(config, validate=_check_config_layers)

    def test_mistral3_architecture_info_can_match_nested_text_model_type(self):
        cfg = SimpleNamespace(
            architectures=["Mistral3ForConditionalGeneration"],
            model_type="mistral3",
            text_config=SimpleNamespace(model_type="ministral3", num_hidden_layers=2),
            vision_config=SimpleNamespace(num_hidden_layers=2),
        )

        arch_info = get_architecture_info(cfg)
        weights = arch_info.all_weights(cfg)
        names = {weight.name for weight in weights}

        assert arch_info.name() == "mistral3"
        assert arch_info.definition.match_model_type == "ministral3"
        assert arch_info.num_layers(cfg) == 2
        assert "vision_tower.transformer.layers.1.attention.q_proj.weight" in names
        assert "language_model.model.layers.1.self_attn.q_proj.weight" in names


class TestSliceConfigMetadata:
    def test_passthrough_slice_uses_selected_layer_types_in_output_config(
        self, tmp_path
    ):
        model_path = tmp_path / "qwen2"
        cfg = Qwen2Config(
            vocab_size=64,
            hidden_size=32,
            intermediate_size=64,
            num_hidden_layers=6,
            num_attention_heads=4,
            num_key_value_heads=4,
            max_position_embeddings=128,
            bos_token_id=1,
            eos_token_id=2,
            pad_token_id=0,
            use_sliding_window=True,
            max_window_layers=2,
        )
        model = Qwen2ForCausalLM(cfg)
        model.save_pretrained(model_path, safe_serialization=True)
        make_tokenizer(vocab_size=64, added_tokens=[]).save_pretrained(model_path)

        merge_config = MergeConfiguration(
            merge_method="passthrough",
            slices=[
                OutputSliceDefinition(
                    sources=[
                        InputSliceDefinition(model=str(model_path), layer_range=[2, 4]),
                    ]
                )
            ],
            dtype="bfloat16",
        )

        arch_info = get_architecture_info(AutoConfig.from_pretrained(model_path))
        out_cfg = _model_out_config(merge_config, arch_info)

        assert out_cfg.num_hidden_layers == 2
        assert out_cfg.layer_types == ["sliding_attention", "sliding_attention"]



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from enum import Enum

import torch


class SparsificationMethod(str, Enum):
    magnitude = "magnitude"
    random = "random"
    magnitude_outliers = "magnitude_outliers"
    rank_magnitude_sampling = "rank_magnitude_sampling"
    consensus_ta = "consensus_ta"
    consensus_ties = "consensus_ties"


def rescale_sum(tensor: torch.Tensor, mask: torch.Tensor):
    """Rescales the values to match the original tensor sum."""
    org_sum = tensor.abs().sum()
    new_sum = (tensor * mask).abs().sum()

    if org_sum >= 1e-8 and new_sum >= 1e-8:
        tensor *= org_sum / new_sum
    return tensor * mask


def magnitude(tensor: torch.Tensor, density: float, rescale: bool) -> torch.Tensor:
    """Masks out the smallest values, retaining a proportion of `density`."""
    if density >= 1:
        return tensor

    k = int(density * tensor.numel())

    assert k > 0, "not gonna zero out the whole tensor buddy"
    mask = torch.zeros_like(tensor)
    w = tensor.abs().view(-1)
    if w.device.type == "cpu":
        w = w.float()
    topk = torch.argsort(w, descending=True)[:k]
    mask.view(-1)[topk] = 1

    if rescale:
        res = rescale_sum(tensor, mask)
    else:
        res = tensor * mask

    return res


def magnitude_outliers(
    tensor: torch.Tensor, density: float, rescale: bool, gamma: float = 0.01
):
    """Masks out smallest values in addition to large outliers.

    The `gamma` proportion of the largest weights are first removed, then the
    smallest weights are removed to achieve the desired density.

    Args:
        tensor (torch.Tensor): The tensor to sparsify.
        density (float): The proportion of weights to retain.
        gamma (float): Percent of largest weights to remove.
    """
    if density >= 1:
        return tensor

    num_elems = tensor.numel()
    target_n = int(density * num_elems)
    n_top = int(gamma * num_elems)
    n_bot = num_elems - target_n - n_top

    if n_bot < 0:
        # cut down on the number of large weights to remove in
        # order to hit the target density
        n_top += n_bot
        n_bot = 0

    w = tensor.abs().view(-1)
    if w.device.type == "cpu":
        w = w.float()
    indices = torch.sort(w, descending=False).indices
    mask = torch.zeros_like(tensor)

    mask.view(-1)[indices[n_bot:-n_top]] = 1

    if rescale:
        res = rescale_sum(tensor, mask)
    else:
        res = tensor * mask
    return res


def bernoulli(tensor: torch.Tensor, density: float, rescale: bool) -> torch.Tensor:
    if density >= 1:
        return tensor

    if (tensor.device.type != "cpu") or tensor.dtype == torch.bfloat16:
        work_dtype = tensor.dtype
    else:
        # torch.bernoulli not implemented for float16 on CPU, upcast to float32
        work_dtype = torch.float32

    mask = torch.bernoulli(
        torch.full_like(input=tensor, fill_value=density, dtype=work_dtype)
    )
    res = tensor.to(work_dtype) * mask
    if rescale:
        res /= density

    return res.to(tensor.dtype)


def rank_magnitude(
    tensor: torch.Tensor, density: float, rescale: bool = True, epsilon: float = 0.05
) -> torch.Tensor:
    if density >= 1:
        return tensor

    if density <= epsilon or density >= (1 - epsilon):
        raise ValueError(
            f"Error: density +- epsilon must be in the range (0, 1). density + epsilon = {density + epsilon}, density - epsilon = {density - epsilon}"
        )

    if (tensor.device.type != "cpu") or tensor.dtype == torch.bfloat16:
        work_dtype = tensor.dtype
    else:
        work_dtype = torch.float32

    if len(tensor.shape) < 2:
        tensor = tensor.unsqueeze(0)

    # Get Rank matrix for the delta values
    tensor_abs = torch.abs(tensor)

    sorted_indices = torch.argsort(tensor_abs, dim=1, descending=False)

    ranking_tensor = torch.zeros_like(tensor_abs, dtype=work_dtype)
    for i in range(tensor_abs.size(0)):
        ranking_tensor[i][sorted_indices[i]] = torch.arange(
            1, tensor.size(1) + 1, dtype=work_dtype
        ).to(tensor.device)

    # Normalise rank matrix to the probability range to density +- epsilon
    range_vals = (
        ranking_tensor.max(dim=1, keepdim=True).values
        - ranking_tensor.min(dim=1, keepdim=True).values
    )
    norm_metrics = (ranking_tensor - ranking_tensor.min(dim=1, keepdim=True).values) / (
        range_vals
    )
    final_probabilities = (density - epsilon) + norm_metrics * (2 * epsilon)

    mask = torch.bernoulli(final_probabilities).to(work_dtype)
    res = tensor.to(work_dtype) * mask

    if rescale:
        res = res / (final_probabilities.to(work_dtype))

    return res.squeeze(0)


def sparsify(
    tensor: torch.Tensor,
    density: float,
    method: SparsificationMethod,
    gamma: float = 0,
    rescale: bool = False,
    epsilon: float = 0.15,
) -> torch.Tensor:
    if (
        method == SparsificationMethod.magnitude
        or method == SparsificationMethod.consensus_ties
    ):
        return magnitude(tensor, density=density, rescale=rescale)
    elif method == SparsificationMethod.random:
        return bernoulli(tensor, density=density, rescale=rescale)
    elif method == SparsificationMethod.magnitude_outliers:
        return magnitude_outliers(tensor, density=density, rescale=rescale, gamma=gamma)
    elif method == SparsificationMethod.rank_magnitude_sampling:
        return rank_magnitude(tensor, density=density, rescale=rescale, epsilon=epsilon)
    else:
        raise NotImplementedError(method)


def get_tall_mask(
    delta: torch.Tensor,  # individual task vectors
    lambda_factor: float,  # hyper-parameter lambda for generating TALL masks
    mixed_delta: torch.Tensor,  # multi-task vector
):
    mask = delta.abs() > lambda_factor * (mixed_delta - delta).abs()
    return mask



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import yaml
from pydantic import BaseModel, model_validator
from typing_extensions import Literal, TypeAlias

from mergekitty.common import ModelReference
from mergekitty.tokenizer.config import TokenizerConfig

ScalarOrGradient: TypeAlias = Union[float, List[float]]


class ConditionalParameter(BaseModel):
    value: ScalarOrGradient
    filter: Optional[str] = None


ParameterSetting: TypeAlias = Union[
    ConditionalParameter, List[ConditionalParameter], ScalarOrGradient
]


def evaluate_setting(
    tensor_name: str, setting: ParameterSetting, t: float = 0
) -> float:
    if isinstance(setting, (float, int, bool, str)):
        return setting
    elif isinstance(setting, list):
        if all(isinstance(e, (int, float)) for e in setting):
            scaled = t * (len(setting) - 1)
            i0 = int(scaled)
            i1 = min(len(setting) - 1, i0 + 1)
            frac = scaled - i0

            return (1 - frac) * setting[i0] + frac * setting[i1]
        elif all(isinstance(e, (float, int, bool, str)) for e in setting):
            return setting[int(t * (len(setting) - 1))]
        else:
            for cond in setting:
                if (
                    (cond.filter is None)
                    or (cond.filter == "*")
                    or (tensor_name and cond.filter in tensor_name)
                ):
                    res = evaluate_setting(tensor_name, cond.value, t)
                    return res
    else:
        raise RuntimeError(f"Unexpected setting value: {setting}")
    return None


class InputSliceDefinition(BaseModel):
    model: ModelReference
    layer_range: Tuple[int, int]
    parameters: Optional[Dict[str, ParameterSetting]] = None


class InputModelDefinition(BaseModel):
    model: ModelReference
    parameters: Optional[Dict[str, ParameterSetting]] = None


class OutputSliceDefinition(BaseModel):
    sources: List[InputSliceDefinition]
    base_model: Optional[ModelReference] = None
    residual_weight: Optional[float] = None
    parameters: Optional[Dict[str, ParameterSetting]] = None


class MergeConfiguration(BaseModel):
    merge_method: str
    slices: Optional[List[OutputSliceDefinition]] = None
    models: Optional[List[InputModelDefinition]] = None
    parameters: Optional[Dict[str, ParameterSetting]] = None
    base_model: Optional[ModelReference] = None
    dtype: Optional[str] = None
    tokenizer_source: Union[Literal["union"], Literal["base"], ModelReference, None] = (
        "base"
    )
    tokenizer: Optional[TokenizerConfig] = None
    chat_template: Optional[str] = None
    out_dtype: Optional[str] = None

    def referenced_models(self) -> List[ModelReference]:
        models = set()
        if self.base_model:
            models.add(self.base_model)
        if self.models:
            for model_in in self.models:
                models.add(model_in.model)
        if self.slices:
            for s in self.slices:
                for src in s.sources:
                    models.add(src.model)
        return list(models)

    @model_validator(mode="after")
    def validate_inputs(self):
        if ((not self.slices) and (not self.models)) or (self.slices and self.models):
            raise RuntimeError("Must specify either output slices or models to merge")
        return self

    @model_validator(mode="after")
    def migrate_and_validate_tokenizer(self):
        if self.tokenizer_source is None and self.tokenizer is None:
            self.tokenizer_source = "base"
        if self.tokenizer_source and self.tokenizer:
            raise RuntimeError("Cannot specify both tokenizer_source and tokenizer")
        return self

    def to_yaml(self) -> str:
        return yaml.dump(
            self.model_dump(exclude_defaults=True, mode="json"),
            Dumper=ConfigYamlDumper,
        ).rstrip()


class ConfigReader(BaseModel):
    config: MergeConfiguration
    t: float
    tensor_name: Optional[str] = None
    slice_out: Optional[OutputSliceDefinition] = None

    @property
    def base_model(self) -> Optional[ModelReference]:
        if self.slice_out and self.slice_out.base_model:
            res = self.slice_out.base_model
        else:
            res = self.config.base_model

        return res

    def for_out_slice(self, slice: OutputSliceDefinition) -> "ConfigReader":
        return ConfigReader(
            config=self.config,
            t=self.t,
            tensor_name=self.tensor_name,
            slice_out=slice,
        )

    def for_tensor(self, tensor_name: str) -> "ConfigReader":
        return ConfigReader(
            config=self.config,
            t=self.t,
            tensor_name=tensor_name,
            slice_out=self.slice_out,
        )

    def with_t(self, t: float) -> "ConfigReader":
        return ConfigReader(
            config=self.config,
            t=t,
            tensor_name=self.tensor_name,
            slice_out=self.slice_out,
        )

    def parameter(
        self,
        name: str,
        model: Optional[ModelReference] = None,
        default: Any = None,
        required: bool = False,
    ) -> Any:
        if self.slice_out:
            if model:
                for s in self.slice_out.sources:
                    if s.model == model and s.parameters and name in s.parameters:
                        value = evaluate_setting(
                            self.tensor_name, s.parameters[name], self.t
                        )
                        if value is not None:
                            return value

            if self.slice_out.parameters and name in self.slice_out.parameters:
                value = evaluate_setting(
                    self.tensor_name, self.slice_out.parameters[name], self.t
                )
                if value is not None:
                    return value

        if self.config.parameters and name in self.config.parameters:
            value = evaluate_setting(
                self.tensor_name,
                self.config.parameters[name],
                self.t,
            )
            if value is not None:
                return value

        if required:
            path_paths = [str(s) for s in [model, self.tensor_name] if s]
            p = ".".join(path_paths)
            suffix = f" for {p}" if p else ""
            raise RuntimeError(f"Missing required parameter {name}{suffix}")
        return default


class ConfigYamlDumper(yaml.Dumper):
    """Custom YAML dumper to format lists of numbers in flow style."""

    def represent_list(self, data: Iterable[Any]) -> yaml.SequenceNode:
        flow_style = all(isinstance(e, (int, float)) for e in data)
        return self.represent_sequence(
            "tag:yaml.org,2002:seq", data, flow_style=flow_style
        )


ConfigYamlDumper.add_representer(list, ConfigYamlDumper.represent_list)



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import importlib
import importlib.resources
import logging
import os
from collections import Counter
from typing import Dict, List, Optional

import tqdm
import transformers

from mergekitty._data import chat_templates
from mergekitty.architecture import (
    ArchitectureInfo,
    get_config_value,
    get_architecture_info,
    set_config_value,
)
from mergekitty.card import generate_card
from mergekitty.config import MergeConfiguration, OutputSliceDefinition
from mergekitty.executor import build_executor
from mergekitty.io.tasks import LoaderCache
from mergekitty.options import MergeOptions
from mergekitty.plan import MergePlanner
from mergekitty.tokenizer import TokenizerInfo


def run_merge(
    merge_config: MergeConfiguration,
    out_path: str,
    options: MergeOptions,
    config_source: Optional[str] = None,
):
    if options.random_seed is not None:
        transformers.trainer_utils.set_seed(options.random_seed)

    if not merge_config.models and not merge_config.slices:
        raise RuntimeError("No output requested")

    model_arch_info = [
        get_architecture_info(m.config(trust_remote_code=options.trust_remote_code))
        for m in merge_config.referenced_models()
    ]
    if not options.allow_crimes:
        if not all(a == model_arch_info[0] for a in model_arch_info[1:]):
            raise RuntimeError(
                "Must specify --allow-crimes to attempt to mix different architectures"
            )
    arch_info = model_arch_info[0]

    # initialize loader cache and set options
    loader_cache = LoaderCache()
    loader_cache.setup(options=options)

    # create config for output model
    cfg_out = _model_out_config(
        merge_config, arch_info, trust_remote_code=options.trust_remote_code
    )

    # warm up loader cache
    for model in (
        pbar := tqdm.tqdm(
            merge_config.referenced_models(),
            desc="Warmup loader cache",
            disable=options.quiet,
        )
    ):
        loader_cache.get(model)
    del pbar

    logging.info("Planning operations")
    targets = MergePlanner(
        merge_config,
        arch_info,
        options=options,
        out_model_config=cfg_out,
    ).plan_to_disk(out_path=out_path)

    exec = build_executor(
        tasks=targets,
        math_device="cuda" if options.cuda else "cpu",
        storage_device="cuda" if options.low_cpu_memory else "cpu",
        executor=options.executor,
    )

    tokenizer = None
    for _task, value in exec.run(quiet=options.quiet):
        if isinstance(value, TokenizerInfo):
            tokenizer = value.tokenizer

    if tokenizer:
        pad_to_multiple_of = None
        if merge_config.tokenizer and merge_config.tokenizer.pad_to_multiple_of:
            pad_to_multiple_of = merge_config.tokenizer.pad_to_multiple_of
        _update_config_vocab(cfg_out, tokenizer, pad_to_multiple_of=pad_to_multiple_of)

    logging.info("Saving config")
    cfg_out.save_pretrained(out_path)

    if options.write_model_card:
        if not config_source:
            config_source = merge_config.to_yaml()

        card_md = generate_card(
            config=merge_config,
            config_yaml=config_source,
            name=os.path.basename(out_path),
        )
        with open(os.path.join(out_path, "README.md"), "w", encoding="utf-8") as fp:
            fp.write(card_md)

        with open(
            os.path.join(out_path, "mergekitty_config.yml"), "w", encoding="utf-8"
        ) as fp:
            fp.write(config_source)

    if tokenizer is None:
        logging.error(
            "Somehow, the tokenizer was not created during the merging process -- did you manually set `tokenizer_source` to None without setting `tokenizer`?"
        )
    else:
        logging.info("Saving tokenizer")
        _set_chat_template(tokenizer, merge_config)
        tokenizer.save_pretrained(out_path, safe_serialization=True)


def _set_chat_template(
    tokenizer: transformers.PreTrainedTokenizerBase,
    merge_config: MergeConfiguration,
    trust_remote_code: bool = False,
):
    chat_template = merge_config.chat_template
    if not chat_template:
        return

    if chat_template == "auto":
        # see if there is a plurality chat template among the input models
        model_templates = []
        for model in merge_config.referenced_models():
            try:
                tok = transformers.AutoTokenizer.from_pretrained(
                    model.model.path,
                    revision=model.model.revision,
                    trust_remote_code=trust_remote_code,
                )
                template = tok.chat_template
                if isinstance(template, dict):
                    template = template.get("default", None)
                if template:
                    model_templates.append(template.strip())
            except Exception as e:
                logging.warning(f"Unable to load tokenizer for {model}", exc_info=e)

        if not model_templates:
            return

        chat_template = Counter(model_templates).most_common(1)[0][0]
        logging.info(f"Auto-selected chat template: {chat_template}")

    elif importlib.resources.is_resource(chat_templates, chat_template + ".jinja"):
        with importlib.resources.open_text(
            chat_templates, chat_template + ".jinja"
        ) as fp:
            chat_template = fp.read()

    elif len(chat_template) < 20 or "{" not in chat_template:
        raise RuntimeError(f"Invalid chat template: {chat_template}")

    tokenizer.chat_template = chat_template


def _model_out_config(
    config: MergeConfiguration,
    arch_info: ArchitectureInfo,
    trust_remote_code: bool = False,
) -> transformers.PretrainedConfig:
    """Return a configuration for the resulting model."""
    if config.base_model:
        res = config.base_model.config(trust_remote_code=trust_remote_code)
    else:
        res = config.referenced_models()[0].config(trust_remote_code=trust_remote_code)
    if config.out_dtype:
        res.torch_dtype = config.out_dtype
    elif config.dtype:
        res.torch_dtype = config.dtype

    if config.slices:
        try:
            num_layers = sum(
                s.sources[0].layer_range[1] - s.sources[0].layer_range[0]
                for s in config.slices
            )
            num_layers_key = arch_info.num_layers_config_key()
            set_config_value(res, num_layers_key, num_layers)
            _resize_layer_types(
                res,
                num_layers_key,
                num_layers,
                config.slices,
                trust_remote_code=trust_remote_code,
            )
        except Exception as e:
            logging.warning(
                "Unable to set number of layers in output config - you may need to manually correct it.",
                exc_info=e,
            )

    return res


def _update_config_vocab(
    config: transformers.PretrainedConfig,
    tokenizer: transformers.PreTrainedTokenizerBase,
    pad_to_multiple_of: Optional[int] = None,
):
    vocab_size = len(tokenizer.get_vocab())
    if pad_to_multiple_of and vocab_size % pad_to_multiple_of:
        vocab_size = vocab_size + pad_to_multiple_of - (vocab_size % pad_to_multiple_of)
    try:
        if hasattr(config, "vocab_size"):
            config.vocab_size = vocab_size
        elif hasattr(config, "text_config") and hasattr(
            config.text_config, "vocab_size"
        ):
            config.text_config.vocab_size = vocab_size
        else:
            raise AttributeError("config has no vocab_size field")
    except Exception as e:
        logging.warning(
            "Unable to set vocabulary size in output config - you may need to manually correct it.",
            exc_info=e,
        )


def _resize_layer_types(
    config: transformers.PretrainedConfig,
    num_layers_key: str,
    num_layers: int,
    slices: List[OutputSliceDefinition],
    trust_remote_code: bool = False,
) -> None:
    layer_types_key = (
        f"{num_layers_key.rsplit('.', 1)[0]}.layer_types"
        if "." in num_layers_key
        else "layer_types"
    )
    try:
        layer_types = get_config_value(config, layer_types_key)
    except (AttributeError, KeyError):
        return

    if not isinstance(layer_types, (list, tuple)) or not layer_types:
        return

    source_config_cache: Dict[str, transformers.PretrainedConfig] = {}
    resized_layer_types = []
    warned_on_mismatch = False
    for out_slice in slices:
        selected_layer_types = None
        for source in out_slice.sources:
            source_key = str(source.model)
            if source_key not in source_config_cache:
                source_config_cache[source_key] = source.model.config(
                    trust_remote_code=trust_remote_code
                )

            try:
                source_layer_types = get_config_value(
                    source_config_cache[source_key], layer_types_key
                )
            except (AttributeError, KeyError):
                return

            if not isinstance(source_layer_types, (list, tuple)):
                return

            start, end = source.layer_range
            layer_type_segment = list(source_layer_types[start:end])
            if len(layer_type_segment) != end - start:
                return

            if selected_layer_types is None:
                selected_layer_types = layer_type_segment
            elif selected_layer_types != layer_type_segment and not warned_on_mismatch:
                logging.warning(
                    "Input slice sources have different layer_types metadata; "
                    "using the first source model values in the output config."
                )
                warned_on_mismatch = True

        if selected_layer_types is not None:
            resized_layer_types.extend(selected_layer_types)

    if len(resized_layer_types) != num_layers:
        return

    set_config_value(config, layer_types_key, resized_layer_types)


__all__ = ["MergeOptions", "run_merge"]



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from typing import Dict, Optional, Union

import pydantic
from pydantic import BaseModel
from typing_extensions import Literal

from mergekitty.common import ModelReference


class ModelTokenEmbedding(BaseModel, frozen=True):
    kind: Literal["model_token"]
    model: ModelReference
    token_id: Optional[int] = None
    token: Optional[str] = None

    @pydantic.model_validator(mode="after")
    def validate_token(self):
        if self.token_id is None and self.token is None:
            raise ValueError("token_id or token must be specified")
        if self.token_id is not None and self.token is not None:
            raise ValueError("only one of token_id or token may be specified")
        return self


class ZeroEmbedding(BaseModel, frozen=True):
    kind: Literal["zero"]


class TokenEmbeddingConfig(BaseModel, frozen=True):
    source: Union[ModelTokenEmbedding, ZeroEmbedding, ModelReference, None] = None
    force: bool = False


class TokenizerConfig(BaseModel, frozen=True):
    source: Union[ModelReference, Literal["union"], Literal["base"]] = "union"
    tokens: Optional[Dict[str, TokenEmbeddingConfig]] = None
    pad_to_multiple_of: Optional[int] = None



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

import torch
from pydantic import BaseModel
from typing_extensions import Literal, override

from mergekitty.architecture import WeightInfo
from mergekitty.common import ImmutableMap, ModelReference
from mergekitty.task import Task
from mergekitty.merge_methods.base import (
    ConfigParameterDef,
    MergeMethod,
    MergeTensorInput,
)
from mergekitty.sparsify import SparsificationMethod, get_tall_mask, sparsify


class ConsensusMethod(str, Enum):
    count = "count"
    sum = "sum"


class GeneralizedTaskArithmeticMerge(MergeMethod, BaseModel, frozen=True):
    consensus_method: Optional[ConsensusMethod]
    sparsification_method: Optional[SparsificationMethod]
    default_normalize: bool
    default_rescale: bool
    method_name: str
    method_pretty_name: Optional[str]
    method_reference_url: Optional[str]

    def name(self) -> str:
        return self.method_name

    @override
    def pretty_name(self) -> Optional[str]:
        return self.method_pretty_name

    @override
    def reference_url(self) -> Optional[str]:
        return self.method_reference_url

    def parameters(self) -> List[ConfigParameterDef]:
        return [
            ConfigParameterDef(name="int8_mask", required=False, default_value=False),
            ConfigParameterDef(
                name="normalize", required=False, default_value=self.default_normalize
            ),
            ConfigParameterDef(
                name="rescale", required=False, default_value=self.default_rescale
            ),
        ]

    def tensor_parameters(self) -> List[ConfigParameterDef]:
        res = [
            ConfigParameterDef(name="weight", required=True),
            ConfigParameterDef(name="density", required=False, default_value=1.0),
        ]
        if self.sparsification_method == SparsificationMethod.magnitude_outliers:
            res.append(
                ConfigParameterDef(
                    name="gamma",
                    default_value=0.01,
                )
            )
        if self.sparsification_method == SparsificationMethod.rank_magnitude_sampling:
            res.append(
                ConfigParameterDef(
                    name="epsilon",
                    default_value=0.15,
                )
            )
            res.append(
                ConfigParameterDef(
                    name="lambda",
                    default_value=1.0,
                )
            )
        if (
            self.sparsification_method == SparsificationMethod.consensus_ta
            or self.sparsification_method == SparsificationMethod.consensus_ties
        ):
            res.append(
                ConfigParameterDef(
                    name="k",
                    default_value=1,
                )
            )
            res.append(
                ConfigParameterDef(
                    name="lambda",
                    default_value=1.0,
                )
            )
        return res

    def make_task(
        self,
        output_weight: WeightInfo,
        tensors: MergeTensorInput,
        base_model: Optional[ModelReference],
        parameters: ImmutableMap[str, Any],
        tensor_parameters: ImmutableMap[ModelReference, ImmutableMap[str, Any]],
    ) -> Task:
        return GTATask(
            method=self,
            tensors=tensors,
            base_model=base_model,
            tensor_parameters=tensor_parameters,
            int8_mask=parameters["int8_mask"],
            normalize=parameters["normalize"],
            rescale=parameters["rescale"],
            weight_info=output_weight,
        )


class GTATask(Task[torch.Tensor]):
    method: GeneralizedTaskArithmeticMerge
    tensors: MergeTensorInput
    base_model: ModelReference
    weight_info: WeightInfo
    tensor_parameters: ImmutableMap[ModelReference, Any]
    int8_mask: bool
    normalize: bool
    rescale: bool

    def uses_accelerator(self) -> bool:
        return True

    def arguments(self) -> Dict[str, Task]:
        return {"tensors": self.tensors}

    def execute(
        self,
        tensors: Dict[ModelReference, torch.Tensor],
        **_kwargs,
    ) -> torch.Tensor:
        # collect task vectors
        tvs, base = get_task_vectors(
            self.weight_info,
            self.base_model,
            tensors,
            tensor_parameters=self.tensor_parameters.data,
        )
        if not tvs:
            return base

        # sparsify
        if (
            self.method.sparsification_method
            and self.method.sparsification_method != SparsificationMethod.consensus_ta
        ):
            for tv_info in tvs:
                kwargs = {}
                if "gamma" in tv_info:
                    kwargs["gamma"] = tv_info["gamma"]

                if "epsilon" in tv_info:
                    kwargs["epsilon"] = tv_info["epsilon"]

                tv_info["sparsified_delta"] = sparsify(
                    tv_info["delta"],
                    density=tv_info["density"],
                    method=self.method.sparsification_method,
                    rescale=self.rescale,
                    **kwargs,
                )

            deltas = torch.stack([tv["sparsified_delta"] for tv in tvs], dim=0)
        else:
            deltas = torch.stack([tv["delta"] for tv in tvs], dim=0)
        weights = torch.tensor(
            [tv["weight"] for tv in tvs], dtype=deltas.dtype, device=deltas.device
        )
        while len(deltas.shape) > len(weights.shape):
            weights.unsqueeze_(-1)

        weighted_deltas = deltas * weights

        # get sign consensus and mix deltas
        if self.method.consensus_method:
            mask_dtype = torch.int8 if self.int8_mask else base.dtype
            mask = get_mask(
                weighted_deltas,
                method=self.method.consensus_method,
                mask_dtype=mask_dtype,
            )
            mixed_delta = (weighted_deltas * mask).sum(dim=0)
            divisor = (weights * mask).sum(dim=0)
            divisor[divisor == 0] = 1
        else:
            mixed_delta = weighted_deltas.sum(dim=0)
            divisor = weights.sum(dim=0)
            divisor[divisor.abs() < 1e-8] = 1

        if self.normalize:
            mixed_delta /= divisor

        if (
            self.method.sparsification_method
            == SparsificationMethod.rank_magnitude_sampling
        ):
            lambda_factor = tvs[0]["lambda"]
            mixed_delta *= lambda_factor

        if (
            self.method.sparsification_method == SparsificationMethod.consensus_ta
            or self.method.sparsification_method == SparsificationMethod.consensus_ties
        ):
            for tv_info in tvs:
                tv_info["tall_mask"] = get_tall_mask(
                    tv_info["delta"],
                    tv_info["lambda"],
                    mixed_delta,
                )
            tall_masks = torch.stack([tv["tall_mask"] for tv in tvs], dim=0)
            consensus_mask = tall_masks.sum(dim=0) >= tvs[0]["k"]
            mixed_delta = mixed_delta * consensus_mask

        return (base + mixed_delta).to(base.dtype)

    def group_label(self) -> Optional[str]:
        return self.tensors.group_label()


def get_task_vectors(
    weight_info: WeightInfo,
    base_model: ModelReference,
    tensors: ImmutableMap[ModelReference, torch.Tensor],
    tensor_parameters: ImmutableMap[ModelReference, ImmutableMap[str, Any]],
) -> Tuple[List[Dict[str, Any]], torch.Tensor]:
    keys = list(tensors.keys())
    base = tensors[base_model]

    parameter_name = weight_info.name

    res = []
    for model in keys:
        if model == base_model:
            continue

        x = tensors[model].to(base.dtype)
        if x.shape != base.shape:
            if weight_info.is_embed:
                x = x[: base.shape[0], : base.shape[1]]
                logging.warning(f"Using submatrix of {model}:{parameter_name}")
            else:
                logging.warning(
                    f"skipping {model}:{parameter_name} due to size mismatch"
                )
                continue

        delta = x - base
        del x
        del tensors[model]

        d = {}
        d["model"] = model
        d["delta"] = delta
        for p in tensor_parameters[model]:
            d[p] = tensor_parameters[model][p]
        res.append(d)
    return res, base


def get_mask(
    delta: torch.Tensor,
    method: Literal["sum", "count"] = "sum",
    mask_dtype: Optional[torch.dtype] = None,
):
    """Returns a mask determining which delta vectors should be merged
    into the final model.

    For the methodology described in the TIES paper use 'sum'. For a
    simpler naive count of signs, use 'count'."""
    if mask_dtype is None:
        mask_dtype = delta.dtype

    sign = delta.sign().to(mask_dtype)

    if method == "sum":
        sign_weight = delta.sum(dim=0)
        majority_sign = (sign_weight >= 0).to(mask_dtype) * 2 - 1
        del sign_weight
    elif method == "count":
        majority_sign = (sign.sum(dim=0) >= 0).to(mask_dtype) * 2 - 1
    else:
        raise RuntimeError(f'Unimplemented mask method "{method}"')

    return sign == majority_sign



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
"""
Module for computational graph execution.

Classes:
    Task: Abstract base class representing a computational task.
"""

from abc import ABC, abstractmethod
from typing import Dict, Optional
from pydantic import BaseModel
from typing_extensions import Generic, TypeVar

ValueT = TypeVar("ValueT")


class Task(ABC, BaseModel, Generic[ValueT], frozen=True, arbitrary_types_allowed=True):
    """
    Abstract base class representing a task in a computational graph.

    This class should be extended to define specific tasks. Each task can have arguments (dependencies) and a defined execution strategy.

    Attributes:
        Generic[ValueT] (TypeVar): The type of the value that the task returns upon execution.

    Methods:
        arguments: Abstract method to define task arguments (dependencies).
        execute: Abstract method to execute the task.
        priority: Returns the priority of the task for scheduling purposes.
        group_label: Returns an optional label for task grouping.
    """

    @abstractmethod
    def arguments(self) -> Dict[str, "Task"]:
        """
        Returns a dictionary of arguments required for this task. The keys of the dictionary
        are argument names, and the values are Task instances. These keys correspond to the
        keyword argument names expected by the execute method.

        For example, if this method returns {'input1': taskA, 'input2': taskB}, the execute
        method should expect to be called as execute(input1=valueA, input2=valueB), where
        valueA and valueB are the outputs of taskA and taskB respectively.

        Returns:
            Dict[str, "Task"]: A dictionary mapping argument names to Task instances.
        """
        ...

    @abstractmethod
    def execute(self, **kwargs) -> ValueT:
        """
        Executes the task using the results of its dependencies.

        The keyword arguments (**kwargs) for this method are dynamically determined based on
        the dictionary returned by the 'arguments' method. Each key in the 'arguments' method's
        return dictionary becomes a keyword argument in this method, with its value being
        the result of the corresponding task's execution.

        Returns:
            ValueT: The result of the task execution.
        """
        ...

    def priority(self) -> int:
        """
        Returns the priority of the task for scheduling.

        Higher numbers indicate higher priority. Default is 0.

        Returns:
            int: The priority of the task.
        """
        return 0

    def group_label(self) -> Optional[str]:
        """
        Returns an optional label used for grouping tasks together.

        Returns:
            Optional[str]: The group label of the task, if any.
        """
        return None

    def uses_accelerator(self) -> bool:
        """
        Returns True if the task can take advantage of matrix operation
        acceleration (such as on a GPU).
        """
        return False



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import random
from typing import List

import click
import yaml

from mergekitty.architecture import get_architecture_info
from mergekitty.common import ModelReference
from mergekitty.config import (
    InputSliceDefinition,
    MergeConfiguration,
    OutputSliceDefinition,
)
from mergekitty.merge import run_merge
from mergekitty.options import MergeOptions, add_merge_options


@click.command("mergekitty-layershuffle")
@click.argument("out_path", type=str)
@click.option("--model", "-m", multiple=True, type=str, help="Add a model to the merge")
@click.option(
    "--weight",
    "-w",
    multiple=True,
    type=float,
    default=[],
    show_default=False,
    help="Weighting for a model",
)
@click.option(
    "--print-yaml/--no-print-yaml",
    is_flag=True,
    help="Print YAML merge config for resulting model",
)
@click.option(
    "--write-yaml",
    type=click.Path(writable=True),
    help="Path to write YAML merge config to",
)
@click.option(
    "--dry-run", is_flag=True, help="Generate a config but do not run the merge"
)
@click.option("--fp16/--no-fp16", is_flag=True, help="Use FP16 precision")
@click.option(
    "--full-random/--no-full-random",
    is_flag=True,
    help="Randomize layer index as well as source model",
)
@add_merge_options
def main(
    out_path: str,
    model: List[str],
    weight: List[float],
    print_yaml: bool,
    write_yaml: bool,
    dry_run: bool,
    fp16: bool,
    full_random: bool,
    merge_options: MergeOptions,
):
    models = [ModelReference.parse(m) for m in model]

    m0_cfg = models[0].config()
    arch_info = get_architecture_info(m0_cfg)
    total_num_layers = arch_info.num_layers(m0_cfg)

    out_slices: List[OutputSliceDefinition] = []

    if full_random:
        for model, frac in zip(models, weight):
            cfg = model.config()
            num_layers = int(arch_info.num_layers(cfg) * frac)
            for _ in range(num_layers):
                src_idx = random.randrange(0, num_layers)
                out_slices.append(
                    OutputSliceDefinition(
                        sources=[
                            InputSliceDefinition(
                                model=str(model),
                                layer_range=(src_idx, src_idx + 1),
                            )
                        ]
                    )
                )
        random.shuffle(out_slices)
    else:
        for layer_idx in range(total_num_layers):
            src_model = random.choices(models, weights=weight, k=1)[0]
            if out_slices and out_slices[-1].sources[0].model == str(src_model):
                out_slices[-1].sources[0].layer_range = (
                    out_slices[-1].sources[0].layer_range[0],
                    layer_idx + 1,
                )
            else:
                out_slices.append(
                    OutputSliceDefinition(
                        sources=[
                            InputSliceDefinition(
                                model=str(src_model),
                                layer_range=(layer_idx, layer_idx + 1),
                            )
                        ]
                    )
                )
    merge_config = MergeConfiguration(
        merge_method="passthrough", slices=out_slices, dtype="float16" if fp16 else None
    )

    if print_yaml or write_yaml:
        yaml_str = yaml.dump(merge_config.model_dump(exclude_none=True, mode="json"))

        if print_yaml:
            print(yaml_str)
        if write_yaml:
            with open(write_yaml, "w", encoding="utf-8") as file:
                file.write(yaml_str)

    if dry_run:
        return

    run_merge(
        merge_config,
        out_path,
        options=merge_options,
    )


if __name__ == "__main__":
    main()



import logging
import os
from typing import Optional

import click
import safetensors.torch
import torch
import tqdm
from transformers import AutoTokenizer

from mergekitty.architecture import get_architecture_info
from mergekitty.common import ModelReference, dtype_from_name
from mergekitty.io.tasks import LoaderCache
from mergekitty.io.tensor_writer import TensorWriter
from mergekitty.options import MergeOptions, add_merge_options


@click.command("mergekitty-activation-based-merge")
@click.argument("model_path", type=str)
@click.argument("secondary_model_path", type=str)
@click.argument("merge_unmerge_directory", type=str)
@click.option("--out-path", "-o", required=True, type=str, help="Output model path")
@click.option(
    "--dtype",
    type=str,
    default="float16",
    help="Data type to convert weights to",
)
@click.option(
    "--device",
    "-d",
    type=str,
    default="cuda",
    help="Device to compute on (default: cuda)",
)
@add_merge_options
def main(
    model_path: str,
    secondary_model_path,
    merge_unmerge_directory: str,
    out_path: str,
    dtype: Optional[str],
    device: Optional[str],
    merge_options: MergeOptions,
):
    model = ModelReference.model_validate(model_path)
    secondary_model = ModelReference.model_validate(secondary_model_path)

    dtype = dtype_from_name(dtype) if dtype else None

    cache = LoaderCache()
    cache.lazy_unpickle = merge_options.lazy_unpickle
    cache.hf_cache_dir = merge_options.transformers_cache

    for m in tqdm.tqdm([model, secondary_model], desc="Preparing models"):
        cache.get(m)

    writer = TensorWriter(
        out_path=out_path,
        max_shard_size=merge_options.out_shard_size,
        safe_serialization=merge_options.safe_serialization,
    )

    model_config = model.config(trust_remote_code=merge_options.trust_remote_code)
    model_arch_info = get_architecture_info(
        model.config(trust_remote_code=merge_options.trust_remote_code)
    )

    loader_1 = cache.get(model)
    loader_2 = cache.get(secondary_model)

    os.makedirs(out_path, exist_ok=True)

    merge_unmerge_dictionary = {}
    # load files from merge_unmerge_directory
    spaces = [
        f.split("_unmerge")[0]
        for f in os.listdir(merge_unmerge_directory)
        if "_unmerge" in f
    ]
    for i in spaces:
        logging.info(f"Loading merge/unmerge tensors for {i}")
        m = safetensors.torch.load_file(
            os.path.join(merge_unmerge_directory, f"{i}_merge.safetensor"),
            device=device,
        )
        u = safetensors.torch.load_file(
            os.path.join(merge_unmerge_directory, f"{i}_unmerge.safetensor"),
            device=device,
        )
        merge_unmerge_dictionary[i] = (
            m[i].to(device, dtype=dtype),
            u[i].to(device, dtype=dtype),
        )

    for weight_info in model_arch_info.all_weights(config=model_config):
        merge_matrix, unmerge_matrix = None, None

        if weight_info.input_space in merge_unmerge_dictionary:
            _, unmerge_matrix = merge_unmerge_dictionary[weight_info.input_space]
            unmerge_matrix = unmerge_matrix.chunk(2, dim=0)

        if weight_info.output_space in merge_unmerge_dictionary:
            merge_matrix, _ = merge_unmerge_dictionary[weight_info.output_space]
            merge_matrix = merge_matrix.chunk(2, dim=1)

        original_w = loader_1.get_tensor(weight_info.name, device=device)
        original_w2 = loader_2.get_tensor(weight_info.name, device=device)

        if dtype is not None:
            original_w = original_w.to(dtype=dtype)
            original_w2 = original_w2.to(dtype=dtype)

        w = torch.clone(original_w)
        w2 = torch.clone(original_w2)

        if not merge_matrix and not unmerge_matrix:
            logging.warning(
                f"❌ Weight {weight_info.name} for model 1 and model 2 has no merge or unmerge matrix"
            )

        if merge_matrix is not None:
            if weight_info.is_embed:
                w = (merge_matrix[0] @ w.T).T
                w2 = (merge_matrix[1] @ w2.T).T
            else:
                w = merge_matrix[0] @ w
                w2 = merge_matrix[1] @ w2

        if unmerge_matrix is not None:
            w = w @ unmerge_matrix[0]
            w2 = w2 @ unmerge_matrix[1]

        # check if weights have not mutated, if yes then  shoot warning
        if torch.allclose(original_w, w):
            logging.warning(
                f"❌ Weight {weight_info.name} for model 1 has NOT mutated during merge"
            )
        else:
            logging.warning(
                f"✅ Weight {weight_info.name} for model 1 has mutated during merge"
            )

        if torch.allclose(original_w2, w2):
            logging.warning(
                f"❌ Weight {weight_info.name} for model 2 has NOT mutated during merge"
            )
        else:
            logging.warning(
                f"✅ Weight {weight_info.name} for model 2 has mutated during merge"
            )

        # average weights and save them
        if merge_matrix:
            w = w + w2
        else:
            w = (w + w2) / 2
        writer.save_tensor(weight_info.name, w)
    writer.finalize()

    tokenizer = AutoTokenizer.from_pretrained(model_path)
    tokenizer.save_pretrained(out_path, safe_serialization=True)

    # write config
    model_out_config = model.config(trust_remote_code=merge_options.trust_remote_code)
    if dtype:
        model_out_config.torch_dtype = dtype
    model_out_config.save_pretrained(out_path)


main()



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import inspect
import typing
from abc import update_abstractmethods
from typing import Any, Dict, List, Optional

import pydantic
import torch
from pydantic import Field
from typing_extensions import Callable

from mergekitty.architecture import WeightInfo
from mergekitty.common import ImmutableMap, ModelReference
from mergekitty.task import Task
from mergekitty.merge_methods.base import (
    ConfigParameterDef,
    MergeMethod,
    MergeTensorInput,
)
from mergekitty.merge_methods.registry import REGISTERED_MERGE_METHODS

STANDARD_KWARGS = {"output_weight", "base_model"}


def __merge_method(
    func: Callable,
    name: str,
    reference_url: Optional[str] = None,
    pretty_name: Optional[str] = None,
) -> Callable:
    use_base_tensor_arg = False
    require_base_tensor = False
    used_kwargs = set()
    parameters: List[ConfigParameterDef] = []
    tensor_parameters: List[ConfigParameterDef] = []
    sig = inspect.signature(func)
    if "tensors" not in sig.parameters:
        raise ValueError("Merge methods must have a 'tensors' parameter")
    tensor_param = sig.parameters["tensors"]
    if (
        (tensor_param.annotation is None)
        or (not hasattr(tensor_param.annotation, "__origin__"))
        or not (
            tensor_param.annotation.__origin__ is list
            and tensor_param.annotation.__args__ == (torch.Tensor,)
        )
    ):
        raise ValueError("'tensors' must be annotated with List[torch.Tensor]")
    if "base_tensor" in sig.parameters:
        bt_param = sig.parameters["base_tensor"]
        if bt_param.annotation == torch.Tensor:
            require_base_tensor = True
        elif (
            hasattr(bt_param.annotation, "__origin__")
            and bt_param.annotation.__origin__ is typing.Union
            and (
                bt_param.annotation.__args__ == (torch.Tensor, type(None))
                or bt_param.annotation.__args__ == (type(None), torch.Tensor)
            )
        ):
            require_base_tensor = False
        else:
            raise ValueError(
                "'base_tensor' must be annotated either torch.Tensor or Optional[torch.Tensor]"
            )
        use_base_tensor_arg = True
    for arg, arg_info in sig.parameters.items():
        if arg in ("base_tensor", "tensors"):
            continue
        if arg in STANDARD_KWARGS:
            used_kwargs.add(arg)
        else:
            if arg_info.annotation is None:
                raise ValueError(
                    "All merge method arguments must have type annotations"
                )
            elif arg_info.annotation in (int, float, bool):
                default_value = arg_info.default
                if default_value == inspect.Parameter.empty:
                    default_value = None
                    required = True
                else:
                    required = False
                parameters.append(
                    ConfigParameterDef(
                        name=arg, required=required, default_value=default_value
                    )
                )
            elif (
                hasattr(arg_info.annotation, "__origin__")
                and arg_info.annotation.__origin__ is list
                and arg_info.annotation.__args__[0] in (float, int)
            ):
                default_value = arg_info.default

                if default_value == inspect.Parameter.empty:
                    default_value = None
                    required = True
                else:
                    required = False
                if (not required) and (not isinstance(default_value, (int, float))):
                    raise ValueError(
                        f"Unexpected default for presumed tensor parameter '{arg}' - should be single number, got {repr(default_value)}"
                    )
                tensor_parameters.append(
                    ConfigParameterDef(
                        name=arg, required=required, default_value=default_value
                    )
                )

    tt_fields = {}
    tt_methods = {}
    tt_fields["gather_tensors"] = (MergeTensorInput, Field(...))
    if ("base_model" in used_kwargs) or use_base_tensor_arg:
        bm_ty = ModelReference if require_base_tensor else Optional[ModelReference]
        field_kwargs = {"default": None} if not require_base_tensor else {}
        tt_fields["base_model"] = (bm_ty, Field(**field_kwargs))
    if "output_weight" in used_kwargs:
        tt_fields["output_weight"] = (WeightInfo, Field(...))
    if parameters:
        tt_fields["parameters"] = (ImmutableMap[str, Any], Field(...))
    if tensor_parameters:
        tt_fields["tensor_parameters"] = (
            ImmutableMap[ModelReference, ImmutableMap[str, Any]],
            Field(...),
        )

    def _arguments(self) -> Dict[str, Task]:
        return {"tensors": self.gather_tensors}

    tt_methods["arguments"] = _arguments

    def _group_label(self) -> Optional[str]:
        return self.gather_tensors.group_label()

    tt_methods["group_label"] = _group_label

    def _uses_accelerator(self) -> bool:
        return True

    tt_methods["uses_accelerator"] = _uses_accelerator

    def _execute(self, tensors: Dict[ModelReference, torch.Tensor], **_kwargs):
        model_refs = set(tensors.keys())
        base_model = getattr(self, "base_model", None)
        if base_model and base_model in model_refs:
            model_refs.remove(base_model)
            if not use_base_tensor_arg:
                model_refs = [base_model] + list(model_refs)
            else:
                model_refs = list(model_refs)
        base_tensor = tensors.get(base_model, None)
        tensors = [tensors[key] for key in model_refs]
        inner_kwargs = {}
        for key in used_kwargs:
            inner_kwargs[key] = getattr(self, key)
        if use_base_tensor_arg:
            inner_kwargs["base_tensor"] = base_tensor
            if require_base_tensor and (inner_kwargs["base_tensor"] is None):
                raise ValueError("Base model tensor required but not present")
        for key in parameters:
            inner_kwargs[key.name] = self.parameters[key.name]
        for key in tensor_parameters:
            inner_kwargs[key.name] = [
                self.tensor_parameters[ref][key.name] for ref in model_refs
            ]
        return func(tensors=tensors, **inner_kwargs)

    tt_methods["execute"] = _execute

    tt_name = f"{name.title().replace(' ', '')}MergeTask"
    tt_cls = pydantic.create_model(tt_name, __base__=Task[torch.Tensor], **tt_fields)
    for method_name, method in tt_methods.items():
        setattr(tt_cls, method_name, method)
    update_abstractmethods(tt_cls)

    mm_fields = {}

    def _make_task(
        self,
        *,
        output_weight: WeightInfo,
        tensors: MergeTensorInput,
        parameters: ImmutableMap[str, Any],
        tensor_parameters: ImmutableMap[ModelReference, ImmutableMap[str, Any]],
        base_model: Optional[ModelReference],
        **_kwargs,
    ) -> Task:
        tt_kwargs = {"gather_tensors": tensors}
        if "base_model" in tt_fields:
            tt_kwargs["base_model"] = base_model
        if "output_weight" in tt_fields:
            tt_kwargs["output_weight"] = output_weight
        if "parameters" in tt_fields:
            tt_kwargs["parameters"] = parameters
        if "tensor_parameters" in tt_fields:
            tt_kwargs["tensor_parameters"] = tensor_parameters
        return tt_cls(**tt_kwargs)

    mm_fields["make_task"] = _make_task

    def _name(self) -> str:
        return name

    mm_fields["name"] = _name

    def _pretty_name(self) -> Optional[str]:
        return pretty_name

    mm_fields["pretty_name"] = _pretty_name

    def _reference_url(self) -> Optional[str]:
        return reference_url

    mm_fields["reference_url"] = _reference_url

    def _tensor_parameters(self) -> List[ConfigParameterDef]:
        return tensor_parameters

    mm_fields["tensor_parameters"] = _tensor_parameters

    def _parameters(self) -> List[ConfigParameterDef]:
        return parameters

    mm_fields["parameters"] = _parameters

    mm_name = f"{name.title().replace(' ', '')}MergeMethod"
    mm_cls = type(mm_name, (MergeMethod,), mm_fields)
    REGISTERED_MERGE_METHODS[name] = mm_cls()
    return func


def merge_method(
    name: str,
    reference_url: Optional[str] = None,
    pretty_name: Optional[str] = None,
) -> Callable:
    """Decorator for registering custom model merging algorithms.

    Enables creation of new merge algorithms that can be specified in merge configurations
    and executed through mergekitty's processing pipeline. Handles parameter validation, task
    creation, and registration in the mergekitty system.

    Args:
        name: Unique identifier for the merge method (lowercase, snake_case recommended)
        reference_url: Optional URL to paper/documentation explaining the method (used in generated READMEs)
        pretty_name: Human-readable display name (used in generated READMEs)

    Returns:
        A decorator that registers the function as a merge method implementation

    Notes:
        The decorated function must meet these requirements:
        - First parameter must be `tensors: List[torch.Tensor]`
        - Must return a single `torch.Tensor`
        - All parameters must have type annotations

        Key behavioral considerations:

        *Base Model Handling:*
        - If the method includes a `base_tensor` parameter:
            * `torch.Tensor` annotation: Requires `base_model` in config, receives its tensor
            * `Optional[torch.Tensor]` annotation: `base_model` optional, `None` if not provided
            * Non-base model tensors passed in `tensors` list
        - Without `base_tensor` parameter:
            * Base model tensor (if specified) will be first in `tensors` list

        *Parameter Types:*
        - Standard parameters (auto-populated):
            * `base_tensor`: Tensor from base model (type determines requirement)
            * `output_weight`: WeightInfo with output configuration
            * `base_model`: ModelReference if using base model logic
        - Scalar parameters (global config):
            * `float`, `int`, or `bool` types specified in top-level `parameters`
        - Tensor parameters (per-model weights):
            * Annotated as `List[float]` or `List[int]`
            * Configured per-model in their `parameters` section
            * Collected into lists ordered by input models

    Example:
        ```python
        @merge_method(
            name="average",
            pretty_name="Simple Average",
            reference_url="https://example.com/mean-merge"
        )
        def average_merge(
            tensors: List[torch.Tensor],  # Input tensors to merge
            weights: List[float],         # Per-model weights (tensor parameter)
            normalize: bool = True        # Scalar parameter
        ) -> torch.Tensor:
            if normalize:
                weights = [w / sum(weights) for w in weights]
            return sum(t * w for t, w in zip(tensors, weights))
        ```

        This would enable merge configurations like:
        ```yaml
        merge_method: average
        parameters:
          normalize: true
        tensor_parameters:
          weights: [0.3, 0.7]
        ```

    Raises:
        ValueError: If function signature doesn't meet requirements
        TypeError: For invalid parameter annotations
    """

    def _wrap(func: Callable) -> Callable:
        return __merge_method(func, name, reference_url, pretty_name)

    return _wrap



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org

from typing import List

from mergekitty.task import Task

from .parallelism import ParallelismExecutor
from .single_threaded import SingleThreadedExecutor


def get_executor_class(name: str):
    normalized = str(name or "single").lower().replace("-", "_")
    if normalized in {"single", "single_threaded"}:
        return SingleThreadedExecutor
    if normalized in {"parallel", "parallelism"}:
        return ParallelismExecutor
    raise ValueError(f"Unknown executor '{name}'")


def build_executor(
    *,
    tasks: List[Task],
    math_device: str,
    storage_device: str,
    executor: str,
):
    executor_cls = get_executor_class(executor)
    return executor_cls(
        tasks=tasks,
        math_device=math_device,
        storage_device=storage_device,
    )


__all__ = [
    "ParallelismExecutor",
    "SingleThreadedExecutor",
    "build_executor",
    "get_executor_class",
]



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

from typing import Any, Dict, List, Optional

import torch

from mergekitty.architecture import WeightInfo
from mergekitty.common import ImmutableMap, ModelReference
from mergekitty.task import Task
from mergekitty.merge_methods.base import (
    ConfigParameterDef,
    MergeMethod,
    MergeTensorInput,
)
from mergekitty.merge_methods.rectify_embed import rectify_embed_sizes


class NearSwapTask(Task[torch.Tensor]):
    gather_tensors: MergeTensorInput
    base_model: ModelReference
    t: float
    weight_info: WeightInfo

    def uses_accelerator(self) -> bool:
        return True

    def arguments(self) -> Dict[str, Task]:
        return {"tensors": self.gather_tensors}

    def execute(self, tensors: Dict[ModelReference, torch.Tensor]) -> torch.Tensor:
        if self.t <= 0:
            raise RuntimeError(f"Threshold cannot be <= zero, got {self.t}")
        if len(tensors) == 1:
            return list(tensors.values())[0]
        elif len(tensors) != 2:
            raise RuntimeError(
                f"Nearswap merge expects exactly two models, got {len(tensors)}"
            )
        elif self.base_model not in tensors:
            raise RuntimeError("Base model not in input tensors")

        [a, b] = list(tensors.items())
        if a[0] != self.base_model:
            [a, b] = [b, a]
        prepped_tensors = [a[1], b[1]]

        rectify_embed_sizes(self.weight_info, prepped_tensors)

        return (
            nearswap(
                self.t,
                prepped_tensors[0],
                prepped_tensors[1],
            )
            .to(prepped_tensors[0].dtype)
            .to(prepped_tensors[0].device)
        )


class NearSwapMerge(MergeMethod):
    def name(self) -> str:
        return "nearswap"

    def pretty_name(self) -> Optional[str]:
        return "NearSwap"

    def reference_url(self) -> Optional[str]:
        return "https://huggingface.co/alchemonaut/QuartetAnemoi-70B-t0.0001"

    def parameters(self) -> List[ConfigParameterDef]:
        return [ConfigParameterDef(name="t", required=True)]

    def make_task(
        self,
        *,
        output_weight: WeightInfo,
        tensors: MergeTensorInput,
        parameters: ImmutableMap[str, Any],
        base_model: Optional[ModelReference],
        **_kwargs,
    ) -> Task:
        return NearSwapTask(
            gather_tensors=tensors,
            base_model=base_model,
            weight_info=output_weight,
            t=parameters["t"],
        )


def nearswap(t: float, v0: torch.Tensor, v1: torch.Tensor) -> torch.Tensor:
    """
    NearSwap implementation using PyTorch.

    Adapted from: https://huggingface.co/alchemonaut/QuartetAnemoi-70B-t0.0001

    Parameters:
        t (float): The sameness threshold.
        v0 (torch.Tensor): Weights from the base model.
        v1 (torch.Tensor): Weights from the secondary model.

    Returns:
        torch.Tensor: Resulting interpolated weights.
    """
    # Compute the absolute difference
    lweight = torch.abs(v0 - v1)

    # Compute the interpolation factor
    lweight = t / lweight
    lweight = torch.nan_to_num(lweight, nan=1.0, posinf=1.0, neginf=1.0)
    lweight = torch.clamp(lweight, min=0.0, max=1.0)

    # Linearly interpolate between v0 and v1
    return lweight * v1 + (1 - lweight) * v0



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import binascii
import logging
import os
import os.path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    Union,
    get_args,
)

import huggingface_hub
import immutables
import peft
import torch
import transformers
from pydantic import BaseModel, model_serializer, model_validator
from pydantic_core import core_schema
from transformers import AutoConfig, PretrainedConfig
from typing_extensions import TypeVar

from mergekitty.io import LazyTensorLoader, ShardedTensorIndex


class ModelPath(BaseModel, frozen=True):
    path: str
    revision: Optional[str] = None

    @model_validator(mode="before")
    def validate_string(cls, value):
        if isinstance(value, str):
            at_ct = value.count("@")
            if at_ct > 1:
                raise RuntimeError(f"Invalid model path - multiple @: {value}")
            elif at_ct == 1:
                path, rev = value.split("@")
                return {"path": path, "revision": rev}
            else:
                return {"path": value}
        return value

    def __str__(self):
        if self.revision:
            return f"{self.path}@{self.revision}"
        return self.path

    def _unique_id(self):
        return (
            os.path.basename(self.path)
            + "_"
            + str(binascii.crc32(self.__str__().encode()))
        )


class ModelReference(BaseModel, frozen=True):
    """A reference to a language model.

    Can be a hf hub path (username/repo), or local. Optionally includes a LoRA."""

    model: ModelPath
    lora: Optional[ModelPath] = None
    override_architecture: Optional[str] = None

    def merged(
        self, cache_dir: Optional[str] = None, trust_remote_code: bool = False
    ) -> "ModelReference":
        """Merge the LoRA if applicable and return a reference to the result."""
        if not self.lora:
            return self

        if not cache_dir:
            raise RuntimeError("Need to specify cache dir to merge adapters")

        out_path = os.path.join(
            cache_dir,
            self.model._unique_id() + "_" + self.lora._unique_id(),
        )

        if not os.path.exists(out_path):
            os.makedirs(out_path, exist_ok=True)

            config = self.config(trust_remote_code)
            auto_cls = _get_auto_cls(config.architectures[0])

            logging.info(f"Loading {self.model} for merge...")
            model = auto_cls.from_pretrained(
                self.model.path,
                revision=self.model.revision,
                torch_dtype=torch.float16,
                low_cpu_mem_usage=True,
                trust_remote_code=trust_remote_code,
            )
            model = peft.PeftModel.from_pretrained(
                model, self.lora.path, revision=self.lora.revision, is_trainable=False
            )
            logging.info(f"Merging {self.lora} into {self.model}")
            model = model.merge_and_unload()
            model.save_pretrained(out_path, safe_serialization=True)
            del model

        return ModelReference(model=out_path)

    def config(self, trust_remote_code: bool = False) -> PretrainedConfig:
        res = AutoConfig.from_pretrained(
            self.model.path,
            revision=self.model.revision,
            trust_remote_code=trust_remote_code,
        )
        if self.override_architecture:
            res.architectures = [self.override_architecture]
        return res

    def tensor_index(self, cache_dir: Optional[str] = None) -> ShardedTensorIndex:
        assert self.lora is None

        path = self.model.path
        if not os.path.exists(path):
            has_safetensors = any(
                fn.lower().endswith(".safetensors")
                for fn in huggingface_hub.list_repo_files(
                    path, repo_type="model", revision=self.model.revision
                )
            )
            patterns = ["tokenizer.model", "*.json"]
            if has_safetensors:
                patterns.append("*.safetensors")
            else:
                patterns.append("*.bin")

            path = huggingface_hub.snapshot_download(
                path,
                revision=self.model.revision,
                cache_dir=cache_dir,
                allow_patterns=patterns,
            )

        return ShardedTensorIndex.from_disk(path)

    def lazy_loader(
        self, cache_dir: Optional[str] = None, lazy_unpickle: bool = True
    ) -> LazyTensorLoader:
        return LazyTensorLoader(
            self.tensor_index(cache_dir),
            lazy_unpickle=lazy_unpickle,
        )

    @model_validator(mode="before")
    def validate_string(cls, value):
        if isinstance(value, str):
            chunks = value.split("+")
            if len(chunks) == 1:
                return {"model": value}
            elif len(chunks) == 2:
                return {"model": chunks[0], "lora": chunks[1]}
            raise RuntimeError(f"Can't parse {value}")
        return value

    @model_serializer()
    def serialize(self):
        res = str(self)
        if '"' in res or " " in res:
            return self
        return res

    @classmethod
    def parse(cls, value: str) -> "ModelReference":
        """Parse a ModelReference. Format: '<MODEL_PATH>(+<LORA_PATH>)?'"""
        return ModelReference.model_validate(value)

    def __str__(self) -> str:
        if self.lora:
            return f"{str(self.model)}+{str(self.lora)}"
        return str(self.model)


class AdapterReference(BaseModel, frozen=True):
    adapter: ModelPath
    override_architecture: Optional[str] = None

    def config(self, trust_remote_code: bool = False) -> PretrainedConfig:
        res = peft.config.PeftConfig.from_pretrained(
            self.adapter.path,
            revision=self.adapter.revision,
            trust_remote_code=trust_remote_code,
        )
        if self.override_architecture:
            res.architectures = [self.override_architecture]
        return res

    def tensor_index(self, cache_dir: Optional[str] = None) -> ShardedTensorIndex:
        path = self.adapter.path
        if not os.path.exists(path):
            has_safetensors = any(
                fn.lower().endswith(".safetensors")
                for fn in huggingface_hub.list_repo_files(
                    path, repo_type="model", revision=self.adapter.revision
                )
            )
            patterns = ["tokenizer.model", "*.json"]
            if has_safetensors:
                patterns.append("*.safetensors")
            else:
                patterns.append("*.bin")

            path = huggingface_hub.snapshot_download(
                path,
                revision=self.adapter.revision,
                cache_dir=cache_dir,
                allow_patterns=patterns,
            )

        return ShardedTensorIndex.from_disk(path)

    def lazy_loader(
        self, cache_dir: Optional[str] = None, lazy_unpickle: bool = True
    ) -> LazyTensorLoader:
        return LazyTensorLoader(
            self.tensor_index(cache_dir),
            lazy_unpickle=lazy_unpickle,
        )

    def __str__(self) -> str:
        return str(self.adapter)


def dtype_from_name(name: Optional[str]) -> Optional[torch.dtype]:
    if not name:
        return None

    if name.startswith("torch."):
        name = name[len("torch.") :]

    if name == "bfloat16":
        return torch.bfloat16
    elif name == "float16":
        return torch.float16
    elif name == "float32":
        return torch.float32
    elif name == "int64":
        return torch.int64
    raise RuntimeError(f'Unimplemented dtype "{name}"')


def parse_kmb(value: Union[str, int]) -> int:
    if isinstance(value, int):
        return value
    elif value.isnumeric():
        return int(value)
    elif value[-1].lower() == "k":
        return int(value[:-1]) * 1000
    elif value[-1].lower() == "m":
        return int(value[:-1]) * 1000 * 1000
    elif value[-1].lower() == "b":
        return int(value[:-1]) * 1000 * 1000 * 1000
    else:
        raise ValueError(value)


T_K = TypeVar("T_K")
T_V = TypeVar("T_V")


class ImmutableMap(Generic[T_K, T_V]):
    data: immutables.Map[T_K, T_V]

    def __init__(self, data: Mapping[T_K, T_V]):
        self.data = data

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Any, handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        instance_schema = core_schema.is_instance_schema(cls)

        args = get_args(source)
        if args:
            dict_schema = handler(Dict[args[0], args[1]])
        else:
            dict_schema = handler(Dict)

        non_instance_schema = core_schema.with_info_after_validator_function(
            lambda value, _info: immutables.Map(value), dict_schema
        )
        return core_schema.union_schema([instance_schema, non_instance_schema])

    def __iter__(self):
        return self.data.__iter__()

    def __getitem__(self, key: T_K) -> T_V:
        return self.data[key]

    def __len__(self) -> int:
        return len(self.data)

    def keys(self) -> Iterator[T_K]:
        return self.data.keys()

    def items(self) -> Iterator[Tuple[T_K, T_V]]:
        return self.data.items()

    def values(self) -> Iterator[T_V]:
        return self.data.values()


def _get_auto_cls(arch_name: str):
    """Get the AutoModel class for a given architecture name."""
    if arch_name.endswith("ForMaskedLM"):
        auto_cls = transformers.AutoModelForMaskedLM
    elif arch_name.endswith("ForConditionalGeneration"):
        auto_cls = getattr(transformers, "AutoModelForImageTextToText", None)
        if auto_cls is None:
            logging.warning(
                "AutoModelForImageTextToText is unavailable - assuming AutoModelForCausalLM"
            )
            auto_cls = transformers.AutoModelForCausalLM
    elif arch_name.endswith("ForSequenceClassification"):
        auto_cls = transformers.AutoModelForSequenceClassification
    elif arch_name.endswith("ForTokenClassification"):
        auto_cls = transformers.AutoModelForTokenClassification
    else:
        if not arch_name.endswith("ForCausalLM") or arch_name.endswith("LMHeadModel"):
            logging.warning(
                f"Unknown model type {arch_name} - assuming AutoModelForCausalLM"
            )
        auto_cls = transformers.AutoModelForCausalLM
    return auto_cls



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import importlib.resources
import string
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field, model_validator
from transformers import PretrainedConfig
from typing_extensions import Literal

import mergekitty._data.architectures


class WeightInfo(BaseModel, frozen=True):
    """Information about an individual weight tensor in a model.

    Attributes:
        name (str):
            The name of the tensor representing the weight.
        is_embed (bool):
            Indicates whether the weight is for an embedding or language model head.
        input_space (Optional[str]):
            The name of the input space associated with the weight, if applicable.
        output_space (Optional[str]):
            The name of the output space associated with the weight, if applicable.
        optional (bool):
            Indicates whether the weight can be omitted from a model.
        aliases (Optional[List[str]]):
            List of alternative names for the weight, if applicable.
        tied_names (Optional[List[str]]):
            List of names for weights that are tied to this weight, if applicable.
        force_dtype (Optional[str]):
            Mandatory dtype for the weight, if applicable.
    """

    name: str
    is_embed: bool = False
    input_space: Optional[str] = None
    output_space: Optional[str] = None
    optional: bool = False
    tied: bool = False
    aliases: Optional[Tuple[str, ...]] = None
    tied_names: Optional[Tuple[str, ...]] = None
    force_dtype: Optional[str] = None
    head_split: Literal[None, "input", "output"] = None
    is_kq: Optional[bool] = False
    is_sparse: Optional[bool] = False


class ProceduralSpaceInfo(BaseModel, frozen=True):
    """Defines a procedural space computed from one or more other spaces.

    Currently only supports residual connections.

    Attributes:
        name (str): The name of the space defined.
        type (str): The type of procedural space.
        inputs (List[str]): List of names of spaces used to define this space."""

    name: str
    type: Literal["residual"]
    inputs: List[str]


class ArchitectureInfo(ABC):
    @abstractmethod
    def name(self) -> str:
        """Return the name of the architecture."""
        ...

    @abstractmethod
    def pre_weights(self, config: PretrainedConfig) -> List[WeightInfo]:
        """Return a list of all weights preceding the first layer."""
        ...

    @abstractmethod
    def post_weights(self, config: PretrainedConfig) -> List[WeightInfo]:
        """Return a list of all weights following the final layer."""
        ...

    @abstractmethod
    def layer_weights(
        self, index: int, config: PretrainedConfig
    ) -> Optional[List[WeightInfo]]:
        """Return a list of all weights associated with a given layer."""
        ...

    @abstractmethod
    def sliceable(self) -> bool:
        """
        Return True if the layers of this architecture can be meaningfully sliced.
        """
        ...

    def num_layers_config_key(self) -> str:
        """Key in config that represents number of layers"""
        return "num_hidden_layers"

    def num_experts_config_key(self) -> str:
        """Key in config that represents number of experts"""
        return "num_experts"

    def num_layers(self, config: PretrainedConfig) -> int:
        """Return the number of layers in a model."""
        return get_config_value(config, self.num_layers_config_key())

    def num_experts(self, config: PretrainedConfig) -> int:
        """Return the number of experts in a model, or None if not applicable."""
        try:
            return get_config_value(config, self.num_experts_config_key())
        except (AttributeError, KeyError):
            return None

    def all_weights(self, config: PretrainedConfig) -> List[WeightInfo]:
        """Return all weights associated with a model."""
        num_layers = self.num_layers(config)
        res = list(self.pre_weights(config))
        for layer_idx in range(num_layers):
            res.extend(self.layer_weights(layer_idx, config))
        res.extend(self.post_weights(config))
        return res

    def procedural_spaces(self, config: PretrainedConfig) -> List[ProceduralSpaceInfo]:
        """Return a list of all procedurally defined spaces in a model."""
        return []

    def has_defined_spaces(self) -> bool:
        """
        Return True if this architecture defines space information needed for
        matching-based merge methods.
        """
        return False


class ConfiguredArchitectureInfo(BaseModel, frozen=True, arbitrary_types_allowed=True):
    info: ArchitectureInfo
    config: PretrainedConfig

    def name(self) -> str:
        return self.info.name()

    def num_layers(self) -> int:
        return self.info.num_layers(self.config)

    def pre_weights(self) -> List[WeightInfo]:
        return self.info.pre_weights(self.config)

    def post_weights(self) -> List[WeightInfo]:
        return self.info.post_weights(self.config)

    def layer_weights(self, index: int) -> List[WeightInfo]:
        return self.info.layer_weights(index, self.config)

    def procedural_spaces(self) -> List[ProceduralSpaceInfo]:
        return self.info.procedural_spaces(self.config)

    def all_weights(self) -> List[WeightInfo]:
        return self.info.all_weights(self.config)


class JSONLayerTemplates(BaseModel, frozen=True):
    weights: List[WeightInfo]
    procedural_spaces: Optional[List[ProceduralSpaceInfo]] = None


class JSONWeightTemplateGroup(BaseModel, frozen=True):
    weights: List[WeightInfo]
    procedural_spaces: Optional[List[ProceduralSpaceInfo]] = None
    count_config_key: Optional[str] = None
    index_name: Optional[str] = None

    @model_validator(mode="after")
    def validate_repeat_fields(self):
        if (self.count_config_key is None) != (self.index_name is None):
            raise ValueError(
                "count_config_key and index_name must be provided together"
            )
        return self


class JSONArchitectureDefinition(BaseModel, frozen=True):
    expected_model_type: str = Field(alias="model_type")
    architectures: List[str]
    match_model_type: Optional[str] = None
    match_model_type_config_key: Optional[str] = None
    pre_weights: List[WeightInfo]
    pre_weight_templates: Optional[List[JSONWeightTemplateGroup]] = None
    layer_templates: JSONLayerTemplates
    post_weights: List[WeightInfo]
    post_weight_templates: Optional[List[JSONWeightTemplateGroup]] = None
    procedural_spaces: Optional[List[ProceduralSpaceInfo]] = None
    num_layers_config_key: Optional[str] = None
    num_experts_config_key: Optional[str] = None

    @model_validator(mode="after")
    def validate_match_model_type_fields(self):
        if (self.match_model_type is None) != (
            self.match_model_type_config_key is None
        ):
            raise ValueError(
                "match_model_type and match_model_type_config_key must be provided together"
            )
        return self


class TemplateWithArithmetic(string.Template):
    idpattern = r"(?a:[_a-z][_a-z0-9]*([+-]1)?)"


def _with_index_substitutions(
    substitutions: Dict[str, int], key: str, value: int
) -> None:
    substitutions[key] = value
    substitutions[f"{key}+1"] = value + 1
    substitutions[f"{key}-1"] = value - 1


def _template_substitution(
    template: str,
    num_layers: int,
    layer_idx: Optional[int] = None,
    num_experts: Optional[int] = None,
    expert_idx: Optional[int] = None,
    extra_substitutions: Optional[Dict[str, int]] = None,
) -> str:
    if "{" not in template:
        return template

    substitutions: Dict[str, int] = {}
    _with_index_substitutions(substitutions, "num_layers", num_layers)

    if layer_idx is not None:
        _with_index_substitutions(substitutions, "layer_index", layer_idx)

    if num_experts is not None:
        _with_index_substitutions(substitutions, "num_experts", num_experts)
        _with_index_substitutions(substitutions, "expert_index", num_experts)
        if expert_idx is not None:
            _with_index_substitutions(substitutions, "expert_index", expert_idx)

    for key, value in (extra_substitutions or {}).items():
        _with_index_substitutions(substitutions, key, value)

    return TemplateWithArithmetic(template).substitute(substitutions)


def get_config_value(config: Union[PretrainedConfig, Dict], key_path: str):
    value = config
    for key in key_path.split("."):
        if isinstance(value, dict):
            value = value[key]
        else:
            value = getattr(value, key)
    return value


def set_config_value(
    config: Union[PretrainedConfig, Dict], key_path: str, value
) -> None:
    keys = key_path.split(".")
    target = config
    for key in keys[:-1]:
        if isinstance(target, dict):
            target = target[key]
        else:
            target = getattr(target, key)

    leaf = keys[-1]
    if isinstance(target, dict):
        target[leaf] = value
    else:
        setattr(target, leaf, value)


class JsonArchitectureInfo(ArchitectureInfo, BaseModel, frozen=True):
    definition: JSONArchitectureDefinition

    def matches_config(self, config: PretrainedConfig) -> bool:
        if self.definition.expected_model_type != config.model_type:
            return False

        if self.definition.match_model_type_config_key is None:
            return True

        try:
            actual_match_model_type = get_config_value(
                config, self.definition.match_model_type_config_key
            )
        except (AttributeError, KeyError):
            return False

        return actual_match_model_type == self.definition.match_model_type

    def _substitute(
        self,
        item: Union[WeightInfo, ProceduralSpaceInfo],
        config: PretrainedConfig,
        layer_idx: Optional[int] = None,
        expert_idx: Optional[int] = None,
        extra_substitutions: Optional[Dict[str, int]] = None,
    ) -> Union[WeightInfo, ProceduralSpaceInfo]:
        num_layers = self.num_layers(config)
        num_experts = self.num_experts(config)

        obj_dict = item.model_dump(mode="json", exclude_unset=True)
        for key in obj_dict:
            if isinstance(obj_dict[key], str):
                obj_dict[key] = _template_substitution(
                    obj_dict[key],
                    num_layers,
                    layer_idx,
                    num_experts,
                    expert_idx,
                    extra_substitutions=extra_substitutions,
                )
            elif isinstance(obj_dict[key], list):
                obj_dict[key] = [
                    (
                        _template_substitution(
                            s,
                            num_layers,
                            layer_idx,
                            num_experts,
                            expert_idx,
                            extra_substitutions=extra_substitutions,
                        )
                        if isinstance(s, str)
                        else s
                    )
                    for s in obj_dict[key]
                ]
        return type(item).model_validate(obj_dict)

    def _substitute_group(
        self,
        group: JSONWeightTemplateGroup,
        config: PretrainedConfig,
        item: Union[WeightInfo, ProceduralSpaceInfo],
    ) -> List[Union[WeightInfo, ProceduralSpaceInfo]]:
        if not group.count_config_key:
            return [self._substitute(item, config=config)]

        repeat_count = get_config_value(config, group.count_config_key)
        return [
            self._substitute(
                item,
                config=config,
                extra_substitutions={group.index_name: index},
            )
            for index in range(repeat_count)
        ]

    def name(self) -> str:
        return self.definition.expected_model_type

    def pre_weights(self, config: PretrainedConfig) -> List[WeightInfo]:
        res = [
            self._substitute(wi, config=config) for wi in self.definition.pre_weights
        ]
        for group in self.definition.pre_weight_templates or []:
            for wi in group.weights:
                res.extend(self._substitute_group(group, config, wi))
        return res

    def layer_weights(
        self, index: int, config: PretrainedConfig
    ) -> Optional[List[WeightInfo]]:
        if self.definition.num_experts_config_key is None:
            return [
                self._substitute(wi, config=config, layer_idx=index)
                for wi in self.definition.layer_templates.weights
            ]
        else:
            expert_weights = [
                wi for wi in self.definition.layer_templates.weights if wi.is_sparse
            ]
            regular_weights = [
                wi for wi in self.definition.layer_templates.weights if not wi.is_sparse
            ]
            return [
                self._substitute(wi, config=config, layer_idx=index)
                for wi in regular_weights
            ] + [
                self._substitute(
                    wi, config=config, layer_idx=index, expert_idx=expert_idx
                )
                for wi in expert_weights
                for expert_idx in range(self.num_experts(config))
            ]

    def post_weights(self, config: PretrainedConfig) -> List[WeightInfo]:
        res = [
            self._substitute(wi, config=config) for wi in self.definition.post_weights
        ]
        for group in self.definition.post_weight_templates or []:
            for wi in group.weights:
                res.extend(self._substitute_group(group, config, wi))
        return res

    def sliceable(self) -> bool:
        return True

    def procedural_spaces(self, config: PretrainedConfig) -> List[ProceduralSpaceInfo]:
        res = []
        for s in self.definition.procedural_spaces or []:
            res.append(self._substitute(s, config=config))
        for group in self.definition.pre_weight_templates or []:
            for s in group.procedural_spaces or []:
                res.extend(self._substitute_group(group, config, s))
        for idx in range(self.num_layers(config)):
            for s in self.definition.layer_templates.procedural_spaces or []:
                res.append(self._substitute(s, config=config, layer_idx=idx))
        for group in self.definition.post_weight_templates or []:
            for s in group.procedural_spaces or []:
                res.extend(self._substitute_group(group, config, s))
        return res

    def has_defined_spaces(self) -> bool:
        if (
            self.definition.procedural_spaces
            or any(
                group.procedural_spaces
                for group in (self.definition.pre_weight_templates or [])
            )
            or self.definition.layer_templates.procedural_spaces
            or any(
                group.procedural_spaces
                for group in (self.definition.post_weight_templates or [])
            )
        ):
            return True
        for wi in (
            self.definition.pre_weights
            + [
                wi
                for group in (self.definition.pre_weight_templates or [])
                for wi in group.weights
            ]
            + self.definition.layer_templates.weights
            + self.definition.post_weights
            + [
                wi
                for group in (self.definition.post_weight_templates or [])
                for wi in group.weights
            ]
        ):
            if wi.input_space or wi.output_space:
                return True
        return False

    def num_layers_config_key(self) -> str:
        return self.definition.num_layers_config_key or super().num_layers_config_key()

    def num_experts_config_key(self) -> str:
        return (
            self.definition.num_experts_config_key or super().num_experts_config_key()
        )


def _load_json_arch(name: str) -> JsonArchitectureInfo:
    text = importlib.resources.read_text(mergekitty._data.architectures, name)
    return JsonArchitectureInfo(
        definition=JSONArchitectureDefinition.model_validate_json(text)
    )


def _load_all_architectures() -> Tuple[
    List[JsonArchitectureInfo], Dict[str, List[JsonArchitectureInfo]]
]:
    architectures: List[JsonArchitectureInfo] = []
    for f in importlib.resources.contents(mergekitty._data.architectures):
        if f.lower().endswith(".json"):
            architectures.append(_load_json_arch(f))

    name_to_arch: Dict[str, List[JsonArchitectureInfo]] = {}
    for arch_info in architectures:
        for name in arch_info.definition.architectures:
            name_to_arch[name] = name_to_arch.get(name, [])
            name_to_arch[name].append(arch_info)
    return architectures, name_to_arch


JSON_ARCHITECTURES, NAME_TO_ARCH = _load_all_architectures()
MISTRAL_INFO = _load_json_arch("mistral.json")
QWEN2_INFO = _load_json_arch("qwen2.json")


def get_architecture_info(config: PretrainedConfig) -> ArchitectureInfo:
    if len(config.architectures) != 1:
        raise RuntimeError("More than one architecture in config?")

    arch_name = config.architectures[0]

    if arch_name not in NAME_TO_ARCH:
        raise RuntimeError(f"Unsupported architecture {arch_name}")

    candidates = list(NAME_TO_ARCH[arch_name])
    matches = [
        candidate for candidate in candidates if candidate.matches_config(config)
    ]
    if len(matches) == 1:
        return matches[0]
    if len(matches) > 1:
        raise RuntimeError(
            f"Ambiguous model_type {config.model_type} for architecture {arch_name}"
        )

    raise RuntimeError(
        f"Unsupported model_type {config.model_type} for architecture {arch_name}"
    )



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
from typing import Dict, Optional

import torch

from mergekitty.common import ImmutableMap, ModelReference
from mergekitty.task import Task
from mergekitty.io.tasks import GatherTensors
from mergekitty.tokenizer.build import BuildTokenizer, TokenizerInfo
from mergekitty.tokenizer.config import (
    ModelTokenEmbedding,
    TokenEmbeddingConfig,
    ZeroEmbedding,
)


class PermutedEmbeddings(Task[Dict[ModelReference, torch.Tensor]]):
    gather_tensors: GatherTensors
    tokenizer_task: BuildTokenizer
    tokens: Optional[ImmutableMap[str, TokenEmbeddingConfig]]
    pad_to_multiple_of: Optional[int]
    base_model: Optional[ModelReference]

    def arguments(self) -> Dict[str, Task]:
        return {"tokenizer_info": self.tokenizer_task, "tensors": self.gather_tensors}

    def execute(
        self, tokenizer_info: TokenizerInfo, tensors: Dict[ModelReference, torch.Tensor]
    ) -> Dict[ModelReference, torch.Tensor]:
        tokenizer = tokenizer_info.tokenizer
        permutations = tokenizer_info.permutations

        models = set(tensors.keys())
        if self.base_model:
            models.add(self.base_model)
        models = list(models)

        vocab = tokenizer.get_vocab()
        vocab_size = len(vocab)
        if self.pad_to_multiple_of and vocab_size % self.pad_to_multiple_of:
            vocab_size = (
                vocab_size // self.pad_to_multiple_of + 1
            ) * self.pad_to_multiple_of
        embed_size = tensors[models[0]].shape[1]
        assert all(t.shape[1] == embed_size for t in tensors.values()), (
            "Embedding sizes must match"
        )

        dtype = tensors[models[0]].dtype
        device = tensors[models[0]].device

        token_configs = dict(**(self.tokens or {}))
        tokens_to_average = self.assign_embedding_sources(
            permutations, models, vocab, token_configs
        )

        default_embeds = {}
        for token, token_id in vocab.items():
            embed = torch.zeros(embed_size, dtype=dtype, device=device)
            if token in tokens_to_average:
                count = 0
                for model in models:
                    p = permutations[model]
                    if p[token_id] < 0:
                        continue
                    embed += tensors[model][p[token_id]]
                    count += 1
                embed /= count
            elif cfg := token_configs.get(token, None):
                cfg: TokenEmbeddingConfig
                embed = self.compute_default_embedding(
                    tokenizer_info, tensors, permutations, token, token_id, cfg
                )
            else:
                continue
            default_embeds[token] = embed

        result = {}
        for model in models:
            p = permutations[model]
            old_embed = tensors[model]
            new_embed = torch.zeros(
                (vocab_size, embed_size), dtype=dtype, device=device
            )
            for token, token_id in vocab.items():
                force = False
                if token in token_configs:
                    force = token_configs[token].force

                if p[token_id] >= 0 and not force:
                    new_embed[token_id, :] = old_embed[p[token_id]]
                elif token in default_embeds:
                    new_embed[token_id, :] = default_embeds[token]
                else:
                    logging.error(
                        f"No embedding for token {repr(token)} in model {model}!"
                    )

            if vocab_size > len(vocab):
                # as suggested by https://nlp.stanford.edu/~johnhew/vocab-expansion.html
                avg_embed = torch.mean(new_embed[: len(vocab), :], dim=0)
                new_embed[len(vocab) :, :] = avg_embed
            result[model] = new_embed

        return result

    def assign_embedding_sources(
        self,
        permutations: Dict[ModelReference, Dict[int, int]],
        models: list[ModelReference],
        vocab: Dict[str, int],
        token_configs: Dict[str, TokenEmbeddingConfig],
    ):
        permutation_list = [permutations[model] for model in models]

        tokens_to_average = set()
        # find tokens that are only present in one model
        for token, token_id in vocab.items():
            if token in token_configs:
                continue

            has_token = [p[token_id] >= 0 for p in permutation_list]
            num_present = sum(int(x) for x in has_token)
            if num_present == 1:
                donor_model = models[has_token.index(True)]
                token_configs[token] = TokenEmbeddingConfig(source=donor_model)
                continue

            if num_present == 0:
                token_configs[token] = TokenEmbeddingConfig(source=ZeroEmbedding())
                logging.warning(f"Token {repr(token)} not found in any model")
                continue

            if num_present > 0 and self.base_model is not None:
                if permutations[self.base_model][token_id] >= 0:
                    token_configs[token] = TokenEmbeddingConfig(source=self.base_model)
                    continue

            tokens_to_average.add(token)
        return tokens_to_average

    def compute_default_embedding(
        self,
        tokenizer_info: TokenizerInfo,
        tensors: Dict[ModelReference, torch.Tensor],
        permutations: Dict[ModelReference, Dict[int, int]],
        token: str,
        token_id: int,
        cfg: TokenEmbeddingConfig,
    ) -> torch.Tensor:
        if isinstance(cfg.source, ZeroEmbedding):
            pass
        elif isinstance(cfg.source, ModelTokenEmbedding):
            model = cfg.source.model
            assert model in permutations, (
                f"Model {model} referenced but not part of merge"
            )
            p = permutations[model]
            src_token_id = cfg.source.token_id
            if src_token_id is None:
                src_token = cfg.source.token
                assert src_token in tokenizer_info.original_vocabs[model], (
                    f"Token {repr(src_token)} not found in model {model}"
                )
                src_token_id = tokenizer_info.original_vocabs[model][src_token]
            assert src_token_id >= 0 and src_token_id < tensors[model].shape[0], (
                f"Token ID {src_token_id} out of range for model {model}"
            )
            embed = tensors[model][src_token_id]
        elif isinstance(cfg.source, ModelReference):
            model = cfg.source
            p = permutations[model]
            assert p[token_id] >= 0, f"Token {repr(token)} not found in model {model}"
            embed = tensors[model][p[token_id]]
        else:
            raise NotImplementedError(cfg)
        return embed



# Partly Copyright (C) 2025 Arcee AI
# Partly Copyright (C) 2025-2026 Allura-org
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import mergekitty.merge_methods.multislerp as multislerp
from mergekitty.merge_methods.base import MergeMethod
from mergekitty.merge_methods.generalized_task_arithmetic import (
    GeneralizedTaskArithmeticMerge,
)
from mergekitty.merge_methods.linear import LinearMerge
from mergekitty.merge_methods.passthrough import PassthroughMerge
from mergekitty.merge_methods.registry import REGISTERED_MERGE_METHODS


def get(method: str) -> MergeMethod:
    if method in REGISTERED_MERGE_METHODS:
        return REGISTERED_MERGE_METHODS[method]
    raise RuntimeError(f"Unimplemented merge method {method}")


__all__ = [
    "MergeMethod",
    "get",
    "LinearMerge",
    "SCEMerge",
    "PassthroughMerge",
    "GeneralizedTaskArithmeticMerge",
    "REGISTERED_MERGE_METHODS",
    "multislerp",
]



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
from typing import Any, Dict, List, Optional, Tuple

import torch
from pydantic import BaseModel
from typing_extensions import override

from mergekitty.architecture import WeightInfo
from mergekitty.common import ImmutableMap, ModelReference
from mergekitty.task import Task
from mergekitty.merge_methods.base import (
    ConfigParameterDef,
    MergeMethod,
    MergeTensorInput,
)


class SCEMerge(MergeMethod, BaseModel, frozen=True):
    def name(self) -> str:
        return "sce"

    @override
    def pretty_name(self) -> str:
        return "SCE"

    @override
    def reference_url(self) -> str:
        return "https://arxiv.org/abs/2408.07990"

    def parameters(self) -> List[ConfigParameterDef]:
        return [
            ConfigParameterDef(name="int8_mask", required=False, default_value=False),
            ConfigParameterDef(name="select_topk", required=False, default_value=1.0),
        ]

    def make_task(
        self,
        output_weight: WeightInfo,
        tensors: MergeTensorInput,
        base_model: Optional[ModelReference],
        parameters: ImmutableMap[str, Any],
        **_kwargs,
    ) -> Task:
        return SCETask(
            tensors=tensors,
            base_model=base_model,
            int8_mask=parameters["int8_mask"],
            select_topk=parameters["select_topk"],
            weight_info=output_weight,
        )


class SCETask(Task[torch.Tensor]):
    tensors: MergeTensorInput
    base_model: ModelReference
    weight_info: WeightInfo
    int8_mask: bool
    select_topk: float

    def uses_accelerator(self) -> bool:
        return True

    def arguments(self) -> Dict[str, Task]:
        return {"tensors": self.tensors}

    def execute(
        self,
        tensors: Dict[ModelReference, torch.Tensor],
        **_kwargs,
    ) -> torch.Tensor:
        # collect task vectors
        tvs, base = get_task_vectors(self.weight_info, self.base_model, tensors)
        if not tvs:
            return base

        deltas = torch.stack([tv["delta"] for tv in tvs], dim=0)
        mask_dtype = torch.int8 if self.int8_mask else base.dtype

        # Select the top τ% elements with high variance
        if self.select_topk < 1:
            mask = get_sce_mask(deltas, self.select_topk, mask_dtype)
            mask_expanded = mask.unsqueeze(0).expand_as(deltas)
            deltas = deltas * mask_expanded

        # Calculate matrix level merging coefficient
        weights = get_sce_weight(deltas)
        weights = torch.tensor(weights, dtype=deltas.dtype, device=deltas.device)
        while len(deltas.shape) > len(weights.shape):
            weights.unsqueeze_(-1)

        # Erase elements with minority directions
        erase_mask = get_erase_mask(
            deltas,
            mask_dtype=mask_dtype,
        )
        erased_weights = weights * erase_mask
        mixed_delta = (deltas * erased_weights).sum(dim=0)

        # Normalize
        divisor = (erased_weights).sum(dim=0)
        divisor[divisor == 0] = 1
        mixed_delta /= divisor

        return (base + mixed_delta).to(base.dtype)


def get_task_vectors(
    weight_info: WeightInfo,
    base_model: ModelReference,
    tensors: ImmutableMap[ModelReference, torch.Tensor],
) -> Tuple[List[Dict[str, Any]], torch.Tensor]:
    keys = list(tensors.keys())
    base = tensors[base_model]

    parameter_name = weight_info.name

    res = []
    for model in keys:
        if model == base_model:
            continue

        x = tensors[model].to(base.dtype)
        if x.shape != base.shape:
            if weight_info.is_embed:
                x = x[: base.shape[0], : base.shape[1]]
                logging.warning(f"Using submatrix of {model}:{parameter_name}")
            else:
                logging.warning(
                    f"skipping {model}:{parameter_name} due to size mismatch"
                )
                continue

        delta = x - base
        del x
        del tensors[model]

        d = {}
        d["model"] = model
        d["delta"] = delta
        res.append(d)
    return res, base


def get_erase_mask(
    delta: torch.Tensor,
    mask_dtype: Optional[torch.dtype] = None,
):
    """Returns a mask determining which delta vectors should be merged
    into the final model.
    """
    if mask_dtype is None:
        mask_dtype = delta.dtype

    sign = delta.sign().to(mask_dtype)

    sign_weight = delta.sum(dim=0)
    majority_sign = (sign_weight >= 0).to(mask_dtype) * 2 - 1
    del sign_weight

    return sign == majority_sign


def get_sce_mask(
    deltas: torch.Tensor,
    density: float,
    mask_dtype: Optional[torch.dtype] = None,
):
    if mask_dtype is None:
        mask_dtype = deltas.dtype
    # Calculate variance along the first dimension
    variance = torch.var(deltas, dim=0, unbiased=False)
    # Count non-zero positions in variance
    non_zero_positions_count = torch.count_nonzero(variance)
    # Calculate the number of top elements to select
    k = int(abs(density) * non_zero_positions_count)
    mask = torch.zeros_like(variance, dtype=mask_dtype)
    if k == 0:
        return mask
    assert k > 0, "not gonna zero out the whole tensor buddy"

    # Get the indices of the top k elements with the highest absolute variance
    topk_indices = torch.topk(variance.abs().view(-1), k=k, largest=True).indices

    mask.view(-1)[topk_indices] = 1
    return mask


def get_sce_weight(deltas):
    # Calculate the squared sum of each delta and normalize by the number of elements
    weights = [torch.sum(delta**2).item() / delta.numel() for delta in deltas]

    # Normalize the weights
    sum_weights = sum(weights)
    if sum_weights == 0:
        weights = [1.0 / len(weights)] * len(weights)
    else:
        weights = [w / sum_weights for w in weights]

    return weights



# Copyright (C) 2025 Arcee AI
#
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.

import logging
from typing import Any, Dict, List, Optional

import torch
from typing_extensions import override

from mergekitty.architecture import WeightInfo
from mergekitty.common import ImmutableMap, ModelReference
from mergekitty.task import Task
from mergekitty.merge_methods.base import (
    ConfigParameterDef,
    MergeMethod,
    MergeTensorInput,
)
from mergekitty.merge_methods.rectify_embed import rectify_embed_sizes


class ModelStockMergeTask(Task[torch.Tensor]):
    gather_tensors: MergeTensorInput
    base_model: ModelReference
    weight_info: WeightInfo
    filter_wise: bool = False

    def uses_accelerator(self) -> bool:
        return True

    def arguments(self) -> Dict[str, Task]:
        return {"tensors": self.gather_tensors}

    def execute(self, tensors: Dict[ModelReference, torch.Tensor]) -> torch.Tensor:
        if len(tensors) == 1 and self.base_model in tensors:
            return tensors[self.base_model]
        if len(tensors) < 3:
            if self.weight_info.optional:
                logging.warning(
                    f"Optional weight {self.weight_info.name} not present in enough models, discarding"
                )
                return None

            raise ValueError(
                "ModelStockMerge requires at least 3 models (base plus two+ others)"
            )

        w_0, ws = self.get_rectified_weights(tensors)
        out_shape = w_0.shape

        if self.filter_wise:
            if w_0.dim() == 1:
                # bias (or other single-vector) parameters should be treated as row vectors
                w_0 = w_0.unsqueeze(0)
                ws = [w.unsqueeze(0) for w in ws]
        else:
            w_0 = w_0.view(-1)
            ws = [w.view(-1) for w in ws]

        offsets = [w - w_0 for w in ws]

        # now there is a question of how to come up with a value for theta.
        # in the two-vector case, we can get an exact angle between the two vectors
        # but the paper doesn't explicitly say what to do in the multi-vector case -
        # they keep using a singular theta value and don't elaborate on how to
        # calculate it. i'm going to assume an average of pairwise angles for now? i guess?

        cos_thetas = []
        for i, w_0_offset in enumerate(offsets):
            for j in range(i + 1, len(offsets)):
                w_1_offset = offsets[j]

                norm_product = torch.norm(w_0_offset, dim=-1) * torch.norm(
                    w_1_offset, dim=-1
                )
                cos_theta = (
                    (w_0_offset * w_1_offset).sum(dim=-1) / norm_product.clamp(min=1e-6)
                ).clamp(-1, 1)
                cos_thetas.append(cos_theta)

        cos_theta = torch.stack(cos_thetas).mean(dim=0).unsqueeze(-1)
        N = len(ws)
        t = (N * cos_theta) / (1 + (N - 1) * cos_theta)

        w_avg = sum(ws) / len(ws)
        w_h = t * w_avg + (1 - t) * w_0

        return w_h.reshape(out_shape)

    def get_rectified_weights(self, tensors: Dict[ModelReference, torch.Tensor]):
        if self.base_model not in tensors:
            raise ValueError("Base model tensor not found")

        all_weights = [tensors[self.base_model]] + [
            tensors[k] for k in tensors if k != self.base_model
        ]
        rectify_embed_sizes(self.weight_info, all_weights)
        w_0 = all_weights[0]
        ws = all_weights[1:]
        return w_0, ws

    def group_label(self) -> Optional[str]:
        return self.gather_tensors.group_label()


class ModelStockMerge(MergeMethod):
    def name(self) -> str:
        return "model_stock"

    @override
    def pretty_name(self) -> Optional[str]:
        return "Model Stock"

    @override
    def reference_url(self):
        return "https://arxiv.org/abs/2403.19522"

    def parameters(self) -> List[ConfigParameterDef]:
        return [
            ConfigParameterDef(name="filter_wise", required=False, default_value=False)
        ]

    def make_task(
        self,
        *,
        output_weight: WeightInfo,
        tensors: MergeTensorInput,
        base_model: Optional[ModelReference],
        parameters: ImmutableMap[str, Any],
        **_kwargs,
    ) -> Task:
        return ModelStockMergeTask(
            gather_tensors=tensors,
            base_model=base_model,
            weight_info=output_weight,
            filter_wise=parameters["filter_wise"],
        )
